#include "Dimension.h"
#include "Constants.h"
#include "Datei.h"
#include "Game.h"

using namespace Framework;


Dimension::Dimension( int id )
    : dimensionId( id ),
    chunks( new Trie<Chunk>() ),
    entities( new RCArray<Entity>() )
{}

Dimension::~Dimension()
{
    entities->release();
    chunks->release();
}

void Dimension::api( Framework::StreamReader *zRequest, NetworkResponse *zResponse )
{
    // TODO: switch type chunck, block, entity
}

void Dimension::tickEntities( Game *zGame )
{
    int index = 0;
    Array<int> removed;
    for( auto entity = entities->getIterator(); entity; entity++, index++ )
    {
        if( zChunk( Punkt( (int)entity->getPosition().x, (int)entity->getPosition().y ) ) )
            entity->tick( this, zGame );
        if( entity->isRemoved() )
            removed.add( index, 0 );
    }
    for( auto i = removed.getIterator(); i; i++ )
        entities->remove( i );
}

void Dimension::getAddrOf( Punkt cPos, char *addr ) const
{
    *(int *)addr = cPos.x;
    *( (int *)addr + 1 ) = cPos.y;
    addr[ 8 ] = 0;
}

void Dimension::getAddrOfWorld( Punkt wPos, char *addr ) const
{
    wPos.x = (int)floor( ( (float)wPos.x + CHUNK_SIZE / 2 ) / CHUNK_SIZE );
    wPos.y = (int)floor( ( (float)wPos.y + CHUNK_SIZE / 2 ) / CHUNK_SIZE );
    getAddrOf( wPos, addr );
}

Chunk *Dimension::zChunk( Punkt wPos ) const
{
    char addr[ 9 ];
    getAddrOfWorld( wPos, addr );
    return chunks->z( addr );
}

Block *Dimension::zBlock( Vec3<int> location )
{
    return zChunk( Punkt( location.x, location.y ) )->zBlockAt( Vec3<int>( ( location.x + CHUNK_SIZE / 2 ) % CHUNK_SIZE, ( location.y + CHUNK_SIZE / 2 ) % CHUNK_SIZE, location.z ) );
}

void Dimension::addEntity( Entity *entity )
{
    entities->add( entity );
}

void Dimension::addChunk( Chunk *chunk )
{
    char addr[ 9 ];
    getAddrOf( chunk->getCenter(), addr );
    if( !chunks->z( addr ) )
    {
        chunks->set( addr, chunk );
        getAddrOf( chunk->getCenter() + Punkt( CHUNK_SIZE, 0 ), addr );
        Chunk *zChunk = chunks->z( addr );
        if( zChunk )
        {
            zChunk->setNeighbor( WEST, chunk );
            chunk->setNeighbor( EAST, chunk );
        }
        getAddrOf( chunk->getCenter() + Punkt( -CHUNK_SIZE, 0 ), addr );
        zChunk = chunks->z( addr );
        if( zChunk )
        {
            zChunk->setNeighbor( EAST, chunk );
            chunk->setNeighbor( WEST, chunk );
        }
        getAddrOf( chunk->getCenter() + Punkt( 0, CHUNK_SIZE ), addr );
        zChunk = chunks->z( addr );
        if( zChunk )
        {
            zChunk->setNeighbor( NORTH, chunk );
            chunk->setNeighbor( SOUTH, chunk );
        }
        getAddrOf( chunk->getCenter() + Punkt( 0, -CHUNK_SIZE ), addr );
        zChunk = chunks->z( addr );
        if( zChunk )
        {
            zChunk->setNeighbor( SOUTH, chunk ); // TODO: correct this in setBlock
            chunk->setNeighbor( NORTH, chunk );
        }
    }
    else
        chunk->release();
}

void Dimension::save( Text worldDir ) const
{
    for( auto chunk = chunks->getIterator(); chunk; chunk++ )
    {
        if( !chunk._ )
            continue;
        Datei *file = new Datei();
        Text filePath = worldDir + "/dim/" + dimensionId + "/";
        filePath.appendHex( chunk->getCenter().x );
        filePath += "_";
        filePath.appendHex( chunk->getCenter().y );
        filePath += ".chunk";
        file->setDatei( filePath );
        if( file->open( Datei::Style::schreiben ) )
            chunk->save( file );
        file->close();
        file->release();
    }
    Text filePath = worldDir + "/dim/" + dimensionId + "/entities";
    Datei *file = new Datei();
    file->setDatei( filePath );
    if( file->open( Datei::Style::schreiben ) )
    {
        for( auto entity = entities->getIterator(); entity; entity++ )
        {
            if( entity->zType()->getId() != PlayerEntityType::ID )
            {
                if( !entity->isRemoved() )
                {
                    int type = entity->zType()->getId();
                    file->schreibe( (char *)&type, 4 );
                    StaticRegistry<EntityType>::INSTANCE.zElement( type )->saveEntity( entity, file );
                }
            }
            else
            {
                Datei pFile;
                pFile.setDatei( worldDir + "/player/" + ( (Player *)entity.val() )->getName() );
                if( pFile.open( Datei::Style::schreiben ) )
                    PlayerEntityType::INSTANCE->saveEntity( entity, &pFile );
            }
        }
        file->close();
    }
}

int Dimension::getDimensionId() const
{
    return dimensionId;
}

bool Dimension::hasChunck( int x, int y ) const
{
    return zChunk( Punkt( x, y ) );
}